/* Regulatory Signal Search (RSS) */
/* Main header file */
/* L.Rubanov, 2006 */
/* 15.04.06  1.25  Remove -1 in MyRandom to eliminate negative answers */
/* 10.04.06  1.24  Enforce portability by RAND_MAX->MYRAND_MAX=32767 */
/* 09.04.06  1.23  Fixed bugs (+1) in history.c and other outputs FOR USER
                   Add U=T equivalence (RSSUTEQIV option)
						 Remove options RSSWAIT, RSSGETS, RSSMPI (no need more)
                   Scale quality etc. to make comparison with RSS02 easier
                   Randomize generator e.g. by clock (add -S<seed> option)
                   Properly associate various outputs with -D and -C levels
                   Review debugging log, also print actual seed value
                   Print reason of algorithm termination
                   -B option added (absolute non-improving iterations limit) */
/* 27.03.06  1.22  Minor changes in console log */
/* 24.03.06  1.21  Console log output added. Several best results output
                   Store and print names of the source sequences in result(s)
                   Arguments -C, -R added. Enforced portability:
                   - own random number generator (by Park & Miller),
                   - eliminate return 0 from qsort() callback functions */
#ifndef RSSHEADER
   #define RSSHEADER
/* Global switches */
/*   #define RSSDEBUG */    /* Enable debugging log to stdout */
	#define RSSUTEQUIV	/* Set U equivalent to T */
   #define RSSPREVENT   /* Prevent repeat use of a permutation */
   #define MYRANDOM		/* Minimal standard portable random number generator */
/* Global constants */
   #define VERSION   "1.25"   /* Program version number */
   #define NMIN      2     /* Minimum number of sequences */
   #define NMAX      256   /* Maximum number of sequences */
   #define LOGNMAX   8     /*    and logarithm NMAX       */
   #define MMAX      1024  /* Maximum length of a sequence */
   #define LMAX      32    /* Maximum length of a word */
   #define LDEF      12    /* Default length of a word */
   #define MAXPATH   256   /* Maximum length of a filename */
   #define DEFOUT    ".out"   /* Extra extension for default output file */
   #define DEFHIST   ".hst"   /* Extra extension for history output file */
   #define COMMENTS  "<>=#@%$()[]-_;,.^\'/"   /* ID's of a comment */
   #define ALPHABET  "actg"   /* Valid data characters */
   #define EMPTY     -1    /* Empty cell of int/short array */
   #define ZERO      (short)mw    /* Denotes 'zero' word */
   #define MAXPVECT  35    /* Max number of permutations generated at a time */
   #define DG_DEF    0     /* Default mode of diagnostics: 0..3 */
                           /* Allowed permutation modes (see pmode below) */
   #define PM_IDENT  1        /* = single (identical) permutation */
   #define PM_CYCLE  2        /* = n cyclic permutations */
   #define PM_FULL   3        /* = n! permutations (brute force) */
   #define PM_RANDOM 4        /* = random selection of next permutation */
   #define PM_ACTIVE 5        /* = active selection of next permutation */
   #define PM_DEF    5        /* Default permutation mode */
   #define LIM_DEF   1000     /* Default max number of permutation */
   #define AG_SUM    1        /* Aggregating by summation */
   #define AG_MAX    2        /* Aggregating by selection of the maximum */
   #define AG_DEF    1        /* Default aggregating mode */
   #define ITER_DEF  5        /* Default number of non-improving iterations */
   #define PV_DEF    12       /* Default limit of new permutations at a time */
   #define MAXNAME	32			/* Max length of sequence name */
   #define RDEF		3			/* Default number of best results to print */
   #define CDEF		1			/* Default console output level */
/* Global types definition */
   typedef unsigned char BYTE;
   typedef struct {        /* Data set for each permutation: */
                  /* Following data allocated only once (initially) */
      short *p;               /* [n]: Current permutation */
      short **tree;           /* [LOGNMAX][nn]: G-tree of sequence numbers */
                     /* Initially (first step) I write data into suu & suv. */
                     /* At each step: suu->suv, then fill suu again. */
      short *suu;             /* [mwz][mwz][nn]: next systems */
      short *suv;             /* [mwz][mwz][nn]: last systems */
      short *l1;              /* [mwz]: words to take from left subtree node */
      short *l2;              /* [mwz]: words to take from right subtree node */
                  /* Following data allocated at the first step (k=logn-1). */
                  /* At subsequent steps: ?uu->?uv and reallocated ?uu*/
      int   *cuu;             /* [mwz][mwz][2**(z-1)]: cost c(u1,u2) */
                  /* The data below is freed at each step except first and */
                  /*  then overwritten by fuu,cuu */
      int   *cuv;             /* [mwz][mwz][2**z]: cost c(u,v) */
   }  SBORKA;
   #define HISTORY struct history
   HISTORY {               /* Joint result list of all used permutations */
      HISTORY *last;          /* pointer to backward */
      HISTORY *next;          /* pointer to forward */
      short *p;               /* [n]: Permutation (NB: n may decrease!) */
      int   pnum;             /* serial number of the permutation */
      int   n, nn, logn;      /* values n, nn, logn for this iteration */
      int   qt;               /* Total quality of the found system (x100) */
      int   qtavg;            /* average proximity in the system (x100) */
      int   power;            /* power of the system (number of n/z words) */
      int   *s;               /* [n0]: the system (words positions per seqs) */
      int   *qw;              /* [n0]: cumulative quality of the words */
      int   *qwavg;           /* [n0]: average proximity of the words (x100) */
   };
   typedef struct {        /* Used to sort the sequences depending on criteria */
      int seqn;               /* native number of a sequence */
      int profile;            /* cumulative quality of a sequence */
      int coverage;           /* cumulative coverage of a sequence */
      int pairs;              /* total number of pairs with a sequence */
   }  SNUM;
/* Global variables and arrays */
   #ifdef RSSDECLARE
      int   n0;               /* initial number of sequences */
      int   nt0;              /* initial number of pairs: n0(n0-1)/2 */
      int   logn0;            /* initial value of logn */
      int   n;                /* number of sequences (NB: decreaseable! ) */
      int   logn;             /* 2**(logn-1) < n <= 2**logn. NB: 2 if n=2 */
      int   nn;               /* 2**logn */
      int   nt;               /* n(n-1)/2 */
      int   m;                /* sequence length (max if different) */
      int   l;                /* word length */
      int   mw;               /* Max words in a seq: m-l (excl. zero word) */
      int   mwz;              /* Max words in a seq: m-l+1 (incl. zero word) */
      int   seqw[NMAX];       /* [NMAX]: number of words in sequences excl. ZW */
      int   threshold;        /* quality threshold (l/2 default) */
      int   first;            /* step counter for Permutation */
      int   debug;            /* write log in output file if nonzero */
      int   pmode;            /* permutation mode (see values above) */
      int   plimit;           /* limit number of permutations */
      char  **seq;            /* array[NMAX] of pointers to input sequences */
      char	**name;				/* array[NMAX] of pointers to sequence names */
      BYTE  **fuv;            /* [nt0][mw][mw]: proximity f(u,v) (triangle) */
      FILE  *infile;          /* Descriptor of input data file */
      FILE  *outfile;         /* Descriptor of result file */
      char  outfilename[MAXPATH+sizeof(DEFOUT)+1];    /* Name of result file */
      HISTORY *hist1;         /* Ptr to first element of historical list */
      HISTORY *hist2;         /* Ptr to last element of historical list */
      int   **lambda;         /* [n0][nw]: current profiles of sequences */
      int   *cover;           /* [nt0]: coverage of sequence pairs (triangle) */
      int   aggr;             /* data aggregation: AG_SUM | AG_MAX */
      short pvect[MAXPVECT+1][NMAX];  /* vector of jointly generated permut-s */
      int   pvmax;            /* Limit of new permutations generated at a time */
      int   qmaximum;         /* global maximum of system quality */
      int   after;            /* number of iterations after last improvement */
      int   aftermax;         /* maximum bound for 'after' in criterion 1 */
      int	afterabs;			/* maximum absolute limit for 'after' */
      int   recursum;         /* total of rejected permutations */
      SNUM  sq[NMAX];         /* Number/quality/coverage of sequences */
      int   pairs;            /* Min pairs per sequence to try (0=all) */
      int   matrix;           /* Prox: 1=take from matrix, 0=calc on the fly */
      int	maxname;				/* Actual max length of sequence name */
      int	rdef;					/* Actual number of best results to print */
      int	cdef;					/* Actual console output level */
      unsigned seed;				/* Seed value for random number generator */
   #else
      extern int n0, nt0, logn0, n, logn, nn, nt, m, l, mw, mwz;
      extern int threshold, first, debug, pmode, plimit, rdef, cdef;
      extern int seqw[NMAX], **lambda, *cover, aggr, pvmax, maxname;
      extern int qmaximum, after, aftermax, afterabs, recursum, pairs, matrix;
      extern short pvect[MAXPVECT][NMAX];
      extern char **seq, **name;
      extern BYTE **fuv;
      extern FILE *infile, *outfile;
      extern char outfilename[MAXPATH+sizeof(DEFOUT)+1];
      extern HISTORY *hist1, *hist2;
      extern SNUM sq[NMAX];
      extern unsigned seed;
   #endif
/* Function prototypes */
   int		AddHistory(SBORKA *sborka);						/* history.c	*/
   int		AnalyzeArgs(int argc, char **argv);				/* analargs.c	*/
   int		AnalyzeHistory(void);					 			/* criteria.c	*/
   int		BestResult(int resnum);		 						/* history.c	*/
   int		Construction(SBORKA *sborka);						/* constr.c		*/
   int		Coverage(SBORKA *sborka);				 			/* history.c	*/
   void		FreeHistory(void);				  					/* freesbrk.c	*/
   void		FreeSborka(SBORKA *sborka);						/* freesbrk.c	*/
   int		GenTree(SBORKA *sborka);							/* gentree.c	*/
   int		MakeBottomLevel(SBORKA *sborka);		  			/* mkbottom.c	*/
   int		MakeNextLevel(SBORKA *sborka, int k); 			/* mknext.c		*/
   int		MakeProximityMatrix(void);		  					/* proxim.c		*/
#ifdef MYRANDOM
	int		MyRandom(void);										/* freesbrk.c	*/
   void		MySrand(unsigned seed);								/* freesbrk.c	*/
#endif
   SBORKA*	Permutation(int *rc);  								/* permut.c		*/
   int		Profile(HISTORY *h);						 			/* history.c	*/
   short		Proximity(int u, int v, int iu, int jv);		/* proxim.c		*/
   int		ReadDataPortion(int *rc);							/* rdata.c		*/
   int		Recurrent(int k);							 			/* criteria.c	*/
   int		SortCoverage(const void *a, const void *b); 	/* criteria.c	*/
   int		SortPairs(const void *a, const void *b);	  	/* criteria.c	*/
   int		SortProfile(const void *a, const void *b);  	/* criteria.c	*/
   int		TryCoverage(void);						 			/* criteria.c	*/
   int		TryProfile(void);							 			/* criteria.c	*/
/* Program completion codes */
   #define OK     0     /* Normal completion */
   #define ERARG  1     /* Erroneous arguments */
   #define ERDATA 2     /* Input data error */
   #define ERIO   4     /* Input/output error */
   #define EODATA 8     /* End of input file */
   #define ERMEM  16    /* Not enough memory or invalid request */
   #define HELP   32    /* Help on arguments requested */
   #define EDEBUG 64    /* Stopped before run-time error to close the log */
   #define ERINT  128   /* Internal error of the algorithm */
#endif

